import * as ko from 'knockout';
let UNINSTANTIABLE_VIEWMODEL_WARNING_ENABLED = true;
const uniqueComponentNames = (function* () {
    let i = 0;
    while (true)
        yield `__router_view_${i++}__`;
})();
// @TODO (try to) make this type-safe, i.e. know it's void if routeConfig.component is void, and middleware if not
export function componentRoutePlugin({ component: componentAccessor, }) {
    if (!componentAccessor)
        return;
    return (ctx) => {
        let isNamedComponent;
        return {
            beforeRender() {
                let _resolve;
                const resolveComponent = (name, instance) => {
                    const c = Object.assign({ name }, (instance || {}));
                    ctx.route.component = name;
                    ctx.component = c;
                    _resolve(c);
                };
                let viewModelInstance;
                ctx.component = new Promise((r) => (_resolve = r));
                ctx.queue(normalizeConfig(ctx, componentAccessor).then(async (normalizedConfig) => {
                    // named component
                    if (typeof normalizedConfig === 'string') {
                        isNamedComponent = true;
                        resolveComponent(normalizedConfig);
                        if (UNINSTANTIABLE_VIEWMODEL_WARNING_ENABLED) {
                            // tslint:disable-next-line:no-console
                            console.warn('[@profiscience/knockout-contrib-router-plugins-component] Unable to instantiate viewModel when using named components. This may cause unexpected behavior. View "Subtleties/Caveats" in the documentation.');
                        }
                    }
                    else {
                        // anonymous component
                        const componentName = normalizedConfig.name || uniqueComponentNames.next().value;
                        const routedComponentInstance = {
                            name: componentName,
                        };
                        const templateConfig = normalizedConfig.template;
                        let viewModelConfig;
                        if (normalizedConfig.viewModel) {
                            try {
                                viewModelInstance = new normalizedConfig.viewModel(ctx);
                                viewModelConfig = { instance: viewModelInstance };
                                routedComponentInstance.viewModel = viewModelInstance;
                            }
                            catch (e) {
                                viewModelConfig = normalizedConfig.viewModel;
                                if (UNINSTANTIABLE_VIEWMODEL_WARNING_ENABLED) {
                                    // tslint:disable-next-line:no-console
                                    console.warn('[@profiscience/knockout-contrib-router-plugins-component] Unable to instantiate viewModel using `new`. This may cause unexpected behavior. See "Subtleties/Caveats" in the documentation.');
                                }
                            }
                        }
                        resolveComponent(componentName, routedComponentInstance);
                        ko.components.register(componentName, {
                            synchronous: true,
                            template: templateConfig,
                            viewModel: viewModelConfig,
                        });
                    }
                }));
            },
            afterRender() {
                if (!isNamedComponent) {
                    ko.components.unregister(ctx.route.component);
                }
            },
            beforeDispose() {
                if (!isNamedComponent && ctx.component) {
                    const c = ctx.component;
                    if (c.viewModel && c.viewModel.dispose) {
                        c.viewModel.dispose();
                        c.viewModel.dispose = () => {
                            /* noop */
                        };
                    }
                }
            },
        };
    };
}
export function disableUninstantiableViewModelWarning() {
    UNINSTANTIABLE_VIEWMODEL_WARNING_ENABLED = false;
}
async function normalizeConfig(ctx, obj) {
    // resolve accessors
    if (typeof obj === 'function')
        obj = obj(ctx);
    // resolve top-level promises
    obj = await obj;
    // top-level default exports
    if (typeof obj.default !== 'undefined')
        obj = obj.default;
    // named components
    if (typeof obj === 'string')
        return obj;
    // anonymous components, may have promised values
    const ret = {};
    await Promise.all(Object.keys(obj).map(async (k) => {
        const v = await obj[k];
        // property default exports
        ret[k] = typeof v.default !== 'undefined' ? v.default : v;
    }));
    return ret;
}
//# sourceMappingURL=index.js.map